@@ -4,7 +4,7 @@
 #
 # License: GPL 3, see file LICENSE
 #
-# Version: 20220629195805
+# Version: 20220709153032
 
 # Based on LocoNet® Personal Use Edition 1.0 SPECIFICATION
 # Which is © Digitrax Inc.
@@ -43,7 +43,7 @@ class Message:
     OPC_LOCO_ADR = 0xBF
     OPC_LOCO_F3 = 0xD4  # not defined in locnet specs, but implemented nevertheless (seen on Roco WLAN maus)
     OPC_SL_RD_DATA = 0xE7
-    OPC_WR_SL_DATA = 0xEF  # not implemented
+    OPC_WR_SL_DATA = 0xEF
 
     def __init__(self, data):
         """
@@ -129,6 +129,8 @@ class Message:
             return SensorState(data)
         elif opcode == 0xB4:
             return LongAcknowledge(data)
+        elif opcode == 0xBA:
+            return MoveSlots(data)
         elif opcode == 0xBB:
             return RequestSlotData(data)
         elif opcode == 0xBC:
@@ -139,6 +141,8 @@ class Message:
             return RequestLocAddress(data)
         elif opcode == 0xE7:
             return SlotDataReturn(data)
+        elif opcode == 0xEF:
+            return WriteSlotData(data)
         return Unknown(data)
 
     @staticmethod
@@ -180,54 +184,136 @@ class Message:
 
         Slots start from zero (but slot 0 is special, as are several others >= 0x70).
         """
-        return (d0 & 0x7F) | ((d1 & 0x0F) << 8)
+        return (d0 & 0x7F) | ((d1 & 0x0F) << 7)
 
 
 class Unknown(Message):
     """
-    An Unknown message simply hold the data bytes.
+    An Unknown message simply holds the data bytes.
     """
 
     pass
 
 
 class PowerOn(Message):
+    """
+    A PowerOn message represents a global track power on message.
+    """
+
     pass
 
 
 class PowerOff(Message):
+    """
+    A PowerOff message represents a global track power off message.
+    """
+
     pass
 
 
 class FunctionGroup1(Message):
-    def __init__(self, data):
-        super().__init__(data)
-        self.slot = int(data[1])
-        self.dir = bool(data[2] & 0x20)
-        self.f0 = bool(data[2] & 0x10)
-        self.f1 = bool(data[2] & 0x1)
-        self.f2 = bool(data[2] & 0x2)
-        self.f3 = bool(data[2] & 0x4)
-        self.f4 = bool(data[2] & 0x8)
+    """
+    A FunctionGroup1 message represents a slot function status change.
+
+    It holds the status for the direction and functions f0 - f4.
+    """
+
+    def __init__(
+        self,
+        data=None,
+        slot=None,
+        dir=None,
+        f0=None,
+        f1=None,
+        f2=None,
+        f3=None,
+        f4=None,
+    ):
+        if data is None:
+            self.slot = slot
+            self.dir = dir
+            self.f0 = f0
+            self.f1 = f1
+            self.f2 = f2
+            self.f3 = f3
+            self.f4 = f4
+            data = bytearray(4)
+            data[0] = 0xA1
+            data[1] = self.slot
+            data[2] = 0
+            data[2] += 0x20 if self.dir else 0
+            data[2] += 0x10 if self.f0 else 0
+            data[2] += 0x1 if self.f1 else 0
+            data[2] += 0x2 if self.f2 else 0
+            data[2] += 0x4 if self.f3 else 0
+            data[2] += 0x8 if self.f4 else 0
+            super().__init__(data)
+            self.updateChecksum()
+        else:
+            if dir is not None or f0 is not None:
+                raise ValueError(
+                    "slot and speed arguments cannot be combined with data argument"
+                )
+            super().__init__(data)
+            self.slot = int(data[1])
+            self.dir = bool(data[2] & 0x20)
+            self.f0 = bool(data[2] & 0x10)
+            self.f1 = bool(data[2] & 0x1)
+            self.f2 = bool(data[2] & 0x2)
+            self.f3 = bool(data[2] & 0x4)
+            self.f4 = bool(data[2] & 0x8)
 
     def __str__(self):
         return f"{self.__class__.__name__}(slot = {self.slot} dir: {self.dir} f0: {self.f0}  f1: {self.f1} f2: {self.f2} f3: {self.f3} f4: {self.f4} | op = {hex(self.opcode)}, {self.length=}, data={list(map(hex,map(int, self.data)))})"
 
 
 class FunctionGroupSound(Message):
-    def __init__(self, data):
-        super().__init__(data)
-        self.slot = int(data[1])
-        self.f5 = bool(data[2] & 0x1)
-        self.f6 = bool(data[2] & 0x2)
-        self.f7 = bool(data[2] & 0x4)
-        self.f8 = bool(data[2] & 0x8)
+    """
+    A FunctionGroupSound message represents a slot function status change.
+
+    It holds the status for functions f5 - f8.
+    """
+
+    def __init__(self, data=None, slot=None, f5=None, f6=None, f7=None, f8=None):
+        if data is None:
+            self.slot = slot
+            self.f5 = f5
+            self.f6 = f6
+            self.f7 = f7
+            self.f8 = f8
+            data = bytearray(4)
+            data[0] = 0xA2
+            data[1] = self.slot
+            data[2] = 0
+            data[2] += 0x1 if self.f5 else 0
+            data[2] += 0x2 if self.f6 else 0
+            data[2] += 0x4 if self.f7 else 0
+            data[2] += 0x8 if self.f8 else 0
+            super().__init__(data)
+            self.updateChecksum()
+        else:
+            if slot is not None or f5 is not None:
+                raise ValueError(
+                    "slot and function arguments cannot be combined with data argument"
+                )
+            super().__init__(data)
+            self.slot = int(data[1])
+            self.f5 = bool(data[2] & 0x1)
+            self.f6 = bool(data[2] & 0x2)
+            self.f7 = bool(data[2] & 0x4)
+            self.f8 = bool(data[2] & 0x8)
 
     def __str__(self):
         return f"{self.__class__.__name__}(slot = {self.slot} f5: {self.f5}  f6: {self.f6} f7: {self.f7} f8: {self.f8} | op = {hex(self.opcode)}, {self.length=}, data={list(map(hex,map(int, self.data)))})"
 
 
 class FunctionGroup2(Message):
+    """
+    A FunctionGroup2 message represents a slot function status change.
+
+    It holds the status for functions f9 - f12.
+    """
+
     def __init__(self, data):
         super().__init__(data)
         self.slot = int(data[1])
@@ -241,6 +327,12 @@ class FunctionGroup2(Message):
 
 
 class FunctionGroup3(Message):
+    """
+    A FunctionGroup3 message represents a slot function status change.
+
+    Depending on the fiegroup, tt holds the status for functions f13 - f19, f21 - f27 or f12 + f20 +f28.
+    """
+
     def __init__(self, data):
         super().__init__(data)
         # data[1] is always 0x20
@@ -279,7 +371,22 @@ class FunctionGroup3(Message):
 
 
 class RequestSwitchFunction(Message):
+    """
+    A RequestSwitchFunction message represents a request for a switch status change.
+
+    It holds the info on whether the switch should be closed or thrown,
+    as well as whether the switch motor should be engaged.
+    """
+
     def __init__(self, data, thrown=None, engage=None):
+        """
+        Construct a RequestSwitchFunction, either from 4 bytes of data or arguments specifying the status for a switch.
+
+        Args:
+            data (bytearray(4) or int): either 4 bytes of raw data or the switch address
+            thrown (bool, optional): if a switch address is given, this should hold thrown or closed. Defaults to None.
+            engage (bool, optional): if a switch address is given, this should signal whether the motor should be engaged . Defaults to None.
+        """ """"""
         if type(data) == int:
             self.address = data
             data = bytearray(4)
@@ -406,11 +513,58 @@ class SlotDataReturn(Message):
         return f"{self.__class__.__name__}(slot={self.slot} loc={self.address} status: {self.status} dir: {self.dir} speed: {self.speed} f0: {self.f0} f1: {self.f1} f2: {self.f2} f3: {self.f3} f4: {self.f4}  f5: {self.f5} f6: {self.f6} f7: {self.f7} f8: {self.f8} trk: {self.trk} ss2: {self.ss2} id1: {self.id1} id2: {self.id2} | op = {hex(self.opcode)}, {self.length=}, data={self.hexdata()})"
 
 
+class WriteSlotData(SlotDataReturn):
+    def __init__(self, slot):
+        if isinstance(slot, bytes) or isinstance(slot, bytearray):
+            super().__init__(data)
+        else:
+            data = bytearray(14)
+            data[0] = 0xEF
+            data[1] = 0x0E
+            data[2] = slot.slot
+            data[3] = slot.status
+            data[4] = slot.address & 0x7F
+            data[9] = slot.address >> 7
+            data[5] = slot.speed
+            data[6] = 0
+            data[6] += 0x20 if slot.dir else 0
+            data[6] += 0x10 if slot.f0 else 0
+            data[6] += 0x1 if slot.f1 else 0
+            data[6] += 0x2 if slot.f2 else 0
+            data[6] += 0x4 if slot.f3 else 0
+            data[6] += 0x8 if slot.f4 else 0
+            data[7] = self.trk
+            data[8] = self.ss2
+            data[10] = 0
+            data[10] += 0x1 if slot.f5 else 0
+            data[10] += 0x2 if slot.f6 else 0
+            data[10] += 0x4 if slot.f7 else 0
+            data[10] += 0x8 if slot.f8 else 0
+            data[11] = self.id1
+            data[12] = self.id2
+            Message.__init__(self, data)  # cannot skip the chain with super()
+            self.updateChecksum()
+
+
 class SlotSpeed(Message):
-    def __init__(self, data):
-        super().__init__(data)
-        self.slot = int(data[1])
-        self.speed = data[2]
+    def __init__(self, data=None, slot=None, speed=None):
+        if data is None:
+            self.slot = slot
+            self.speed = speed
+            data = bytearray(4)
+            data[0] = 0xA0
+            data[1] = slot
+            data[2] = speed
+            super().__init__(data)
+            self.updateChecksum()
+        else:
+            if slot is not None or speed is not None:
+                raise ValueError(
+                    "slot and speed arguments cannot be combined with data argument"
+                )
+            self.slot = int(data[1])
+            self.speed = data[2]
+            super().__init__(data)
 
     def __str__(self):
         return f"{self.__class__.__name__}(slot={self.slot} speed: {self.speed} | op = {hex(self.opcode)}, {self.length=}, data={self.hexdata()})"
@@ -434,6 +588,28 @@ class RequestLocAddress(Message):
         return f"{self.__class__.__name__}(address = {self.address} | op = {hex(self.opcode)}, {self.length=}, data={self.hexdata()})"
 
 
+class MoveSlots(Message):
+    def __init__(self, data=None, src=None, dst=None):
+        if data is None:
+            self.src = src
+            self.dst = dst
+            data = bytearray(4)
+            data[0] = 0xBA
+            data[1] = self.src
+            data[2] = self.dst
+            super().__init__(data)
+            self.updateChecksum()
+        else:
+            if src is not None or dst is not None:
+                raise ValueError("slot arguments cannot be combined with data argument")
+            super().__init__(data)
+            self.src = int(data[1])
+            self.dst = int(data[2])
+
+    def __str__(self):
+        return f"{self.__class__.__name__}(src = {self.src} dst = {self.dst}| op = {hex(self.opcode)}, {self.length=}, data={self.hexdata()})"
+
+
 class CaptureTimeStamp(Message):
     def __init__(self, t):
         if isinstance(t, time):
