@@ -4,16 +4,11 @@
 #
 # License: GPL 3, see file LICENSE
 #
-# Version: 20220626170320
+# Version: 20220710113627
 
-import argparse
-import sys
-import time
 from threading import Thread
 from time import sleep
 
-from .Interface import Interface
-from .Scrollkeeper import Scrollkeeper
 from .Utils import Args, createInterface, createScrollkeeper, reporter
 
 if __name__ == "__main__":
